#include "burp.h"
#include "prog.h"
#include "msg.h"
#include "lock.h"
#include "rs_buf.h"
#include "handy.h"
#include "asyncio.h"
#include "zlibio.h"
#include "counter.h"

#include <netdb.h>
#include <librsync.h>

static void reuseaddr(int fd)
{
	int tmpfd;
	setsockopt(fd, SOL_SOCKET, SO_REUSEADDR,
		(sockopt_val_t)&tmpfd, sizeof(tmpfd));
}

static void sighandler(int sig)
{
	logp("got signal: %d\n", sig);
	// Close the sockets properly so as to avoid annoying waits during
	// testing when I kill the server with a Ctrl-C and then get
	// 'unable to bind listening socket'.
	async_free();
	logp("exiting\n");
	exit(1);
}

static int init_listen_socket(int port)
{
	int rfd;
	struct sockaddr_in laddr;
	socklen_t laddrlen;

	if((rfd=socket(AF_INET, SOCK_STREAM, IPPROTO_TCP))<0)
	{
		logp("%s: unable to create listening socket\n");
		return -1;
	}

	laddr.sin_family=AF_INET;
	laddr.sin_port=htons(port);
	laddr.sin_addr.s_addr=INADDR_ANY;

	// Bind listening address.
	laddrlen=sizeof(laddr);

	if(bind(rfd, (struct sockaddr *)&laddr, laddrlen)<0)
	{
		logp("%s: unable to bind listening socket\n");
		return -1;
	}

	reuseaddr(rfd);

	getsockname(rfd, (struct sockaddr *)&laddr, &laddrlen);

	// Say that we are happy to accept connections.
	if(listen(rfd, 5)<0)
	{
		close_fd(&rfd);
		logp("could not listen on main socket\n");
		return -1;
	}

#ifdef HAVE_WIN32
	{
		u_long ioctlArg=0;
		ioctlsocket(rfd, FIONBIO, &ioctlArg);
	}
#endif


	return rfd;
}

static void setup_signals(void)
{
	// Ignore SIGPIPE - we are careful with read and write return values.
#ifndef HAVE_WIN32
	signal(SIGPIPE, SIG_IGN);
	// Get rid of defunct children.
	signal(SIGCHLD, SIG_IGN);

	signal(SIGABRT, &sighandler);
	signal(SIGTERM, &sighandler);
	signal(SIGINT, &sighandler);
#endif
}

static int recursive_hardlink(const char *src, const char *dst)
{
	int n=-1;
	int ret=0;
	struct dirent **dir;
	char *tmp=NULL;
	logp("in rec hl: %s %s\n", src, dst);
	if(!(tmp=prepend_s(dst, "dummy", strlen("dummy"))))
		return -1;
	if(mkpath(&tmp))
	{
		logp("could not mkpath for %s\n", tmp);
		free(tmp);
		return -1;
	}
	free(tmp);

	if((n=scandir(src, &dir, 0, 0))<0)
	{
		logp("recursive_hardlink scandir %s: %s\n",
			src, strerror(errno));
		return -1;
	}
	while(n--)
	{
		char *fullpatha=NULL;
		char *fullpathb=NULL;
		if(dir[n]->d_ino==0
		  || !strcmp(dir[n]->d_name, ".")
		  || !strcmp(dir[n]->d_name, ".."))
			{ free(dir[n]); continue; }
		if(!(fullpatha=prepend_s(src,
			dir[n]->d_name, strlen(dir[n]->d_name)))
		|| !(fullpathb=prepend_s(dst,
			dir[n]->d_name, strlen(dir[n]->d_name))))
		{
			if(fullpatha) free(fullpatha);
			if(fullpathb) free(fullpathb);
			break;
		}

		if(is_dir(fullpatha))
		{
			if(recursive_hardlink(fullpatha, fullpathb))
			{
				free(fullpatha);
				free(fullpathb);
				break;
			}
		}
		else
		{
			logp("hardlinking %s to %s\n", fullpathb, fullpatha);
			if(link(fullpatha, fullpathb))
			{
				logp("hardlink %s to %s failed: %s\n",
					fullpathb, fullpatha, strerror(errno));
				free(fullpatha);
				free(fullpathb);
				break;
			}
		}
		free(fullpatha);
		free(fullpathb);
		free(dir[n]);
	}
	if(n>0)
	{
		ret=-1;
		for(; n>0; n--) free(dir[n]);
	}
	free(dir);

	return ret;
}

#define RECDEL_ERROR			-1
#define RECDEL_OK			0
#define RECDEL_ENTRIES_REMAINING	1

static int recursive_delete(const char *d, const char *file, bool delfiles)
{
	int n=-1;
	int ret=RECDEL_OK;
	struct dirent **dir;
	struct stat statp;
	char *directory=NULL;

	if(!file)
	{
		if(!(directory=prepend_s(d, "", 0))) return RECDEL_ERROR;
	}
	else if(!(directory=prepend_s(d, file, strlen(file))))
	{
		logp("out of memory\n");
		return RECDEL_ERROR;
	}

	if(lstat(directory, &statp))
	{
		// path does not exist.
		free(directory);
		return RECDEL_OK;
	}

	if((n=scandir(directory, &dir, 0, 0))<0)
	{
		logp("scandir %s: %s\n", directory, strerror(errno));
		free(directory);
		return RECDEL_ERROR;
	}
	while(n--)
	{
		char *fullpath=NULL;
		if(dir[n]->d_ino==0
		  || !strcmp(dir[n]->d_name, ".")
		  || !strcmp(dir[n]->d_name, ".."))
			{ free(dir[n]); continue; }
		if(!(fullpath=prepend_s(directory,
			dir[n]->d_name, strlen(dir[n]->d_name))))
				break;

		if(is_dir(fullpath))
		{
			int r;
			if((r=recursive_delete(directory,
				dir[n]->d_name, delfiles))==RECDEL_ERROR)
			{
				free(fullpath);
				break;
			}
			// do not overwrite ret with OK if it previously
			// had ENTRIES_REMAINING
			if(r==RECDEL_ENTRIES_REMAINING) ret=r;
		}
		else if(delfiles)
		{
			if(unlink(fullpath))
			{
				logp("unlink %s: %s\n",
					fullpath, strerror(errno));
				ret=RECDEL_ENTRIES_REMAINING;
			}
		}
		else
		{
			ret=RECDEL_ENTRIES_REMAINING;
		}
		free(fullpath);
		free(dir[n]);
	}
	if(n>0)
	{
		ret=RECDEL_ERROR;
		for(; n>0; n--) free(dir[n]);
	}
	free(dir);

	if(ret==RECDEL_OK && rmdir(directory))
	{
		logp("rmdir %s: %s\n", directory, strerror(errno));
		ret=RECDEL_ERROR;
	}
	free(directory);
	return ret;
}

/*
#ifndef HAVE_WIN32
static int make_link(const char *fname, size_t flen, const char *lnk, size_t llen, char cmd, const char *datadir)
{
	int ret=0;
	char *lpath=NULL;

	if(cmd=='L')
	{
		if(!(lpath=prepend_s(datadir, lnk, llen))) return -1;
		printf("%s -> %s\n", fname, lpath);
		ret=link(lpath, fname);
	}
	else
	{
		ret=symlink(lnk, fname);
		printf("%s -> %s\n", fname, lnk);
	}

	if(ret) logp("could not %slink %s -> %s\n", cmd=='L'?"hard":"sym",
		fname, cmd=='L'?lpath:fname);
	if(lpath) free(lpath);

	return ret;
}
#endif
*/

static int check_client_and_password(struct config *conf, const char *client, const char *password, struct config *cconf)
{
	char cpath[256]="";
	snprintf(cpath, sizeof(cpath), "%s/%s", conf->clientconfdir, client);

	// Some client settings can be globally set in the server config and
	// overridden in the client specific config.
	init_config(cconf);
	if(set_client_global_config(conf, cconf)
	  || load_config(cpath, cconf, 0))
		return -1;

	if(!cconf->password || strcmp(cconf->password, password))
	{
		logp("password rejected for client %s\n", client);
		return -1;
	}
	if(!cconf->keep)
	{
		logp("%s: you cannot set the keep value for a client to 0!\n",
				client);
		return -1;
	}
	return 0;
}

static int authorise_client(struct config *conf, char **client, struct config *cconf)
{
	char cmd;
	size_t len=0;
	char *buf=NULL;
	char *password=NULL;

	if(async_read(&cmd, &buf, &len))
	{
		logp("unable to read initial message\n");
		return -1;
	}
	if(cmd!='c' || strcmp(buf, "hello"))
	{
		logp("unexpected command given: %c %s\n", cmd, buf);
		free(buf);
		return -1;
	}
	free(buf); buf=NULL;
	async_write_str('c', "whoareyou");
	if(async_read(&cmd, &buf, &len) || !len)
	{
		logp("unable to get client name\n");
		return -1;
	}
	*client=buf;
	buf=NULL;
	async_write_str('c', "okpassword");
	if(async_read(&cmd, &buf, &len) || !len)
	{
		logp("unable to get password for client %s\n", *client);
		free(*client);
		free(buf); buf=NULL;
		return -1;
	}
	password=buf;
	buf=NULL;

	if(check_client_and_password(conf, *client, password, cconf))
	{
		free(*client);
		free(password);
		return -1;
	}

	async_write_str('c', "ok");
	logp("auth ok for client: %s\n", *client);
	free(password);
	return 0;
}

static int do_patch(const char *dst, const char *del, const char *upd, bool gzupd)
{
	FILE *dstp=NULL;
	gzFile delp=NULL;
	gzFile updp=NULL;
	FILE *updfp=NULL;
	rs_result result;

	logp("patching...\n");

	if(!(dstp=fopen(dst, "rb")))
	{
		logp("could not open %s for reading\n", dst);
		return -1;
	}
	else if(!(delp=gzopen(del, "rb9")))
	{
		logp("could not open %s for reading\n", del);
		close_fp(&dstp);
		return -1;
	}
	else if(gzupd)
		updp=gzopen(upd, "wb9");
	else
		updfp=fopen(upd, "wb");

	if(!updp && !updfp)
	{
		logp("could not open %s for writing\n", upd);
		close_fp(&dstp);
		gzclose_fp(&delp);
		return -1;
	}
	

	result=rs_patch_gzfile(dstp, delp, updfp, updp, NULL);
	fclose(dstp);
	gzclose_fp(&delp);
	if(updp) gzclose_fp(&updp);
	if(updfp) fclose(updfp);

	return result;
}

static int receive_dir(char cmd, const char *dir, size_t len, FILE *mp, const char *statbuf, size_t slen)
{
	send_msg_fp(mp, 'r', statbuf, slen);
	send_msg_fp(mp, cmd, dir, len);
	return 0;
}

static int receive_special(char cmd, const char *dir, size_t len, FILE *mp, const char *statbuf, size_t slen)
{
	send_msg_fp(mp, 'r', statbuf, slen);
	send_msg_fp(mp, cmd, dir, len);
	return 0;
}

static int receive_link(char cmd, const char *fname, size_t flen, FILE *mp, const char *statbuf, size_t slen)
{
	int ret=0;
	char lcmd;
	size_t llen=0;
	char *lname=NULL;

	// Read where the link points to.
	if(async_read(&lcmd, &lname, &llen))
	{
		ret=-1;
	}
	else
	{
		send_msg_fp(mp, 'r', statbuf, slen);
		send_msg_fp(mp, cmd, fname, flen);
		send_msg_fp(mp, lcmd, lname, llen);
	}
	if(lname) free(lname);

	return ret;
}

struct dpth
{
	int prim;
	int seco;
	int tert;
	char path[32];
	int looped;
};

static void mk_dpth(struct dpth *dpth)
{
	snprintf(dpth->path, sizeof(dpth->path), "%04X/%04X/%04X.gz",
		dpth->prim, dpth->seco, dpth->tert);
}

static void mk_dpth_prim(struct dpth *dpth)
{
	snprintf(dpth->path, sizeof(dpth->path), "%04X", dpth->prim);
}

static void mk_dpth_seco(struct dpth *dpth)
{
	snprintf(dpth->path, sizeof(dpth->path), "%04X/%04X",
		dpth->prim, dpth->seco);
}

static int get_highest_entry(const char *path)
{
	int ent=0;
	int max=0;
	DIR *d=NULL;
	struct dirent *dp=NULL;

	if(!(d=opendir(path))) return -1;
	while((dp=readdir(d)))
	{
		if(dp->d_ino==0
		  || !strcmp(dp->d_name, ".")
		  || !strcmp(dp->d_name, ".."))
			continue;
		ent=strtol(dp->d_name, NULL, 16);
		if(ent>max) max=ent;
	}
	closedir(d);
	return max;
}

// Three levels with 65535 entries each gives
// 65535^3 = 281,462,092,005,375 data entries
static int incr_dpth(struct dpth *dpth)
{
	if(dpth->tert++>=0xFFFF)
	{
		dpth->tert=0;
		if(dpth->seco++>=0xFFFF)
		{
			dpth->seco=0;
			if(dpth->prim++>=0xFFFF)
			{
				dpth->prim=0;
				// Start again from zero, so make sure that
				// the initial open of a data file is in an
				// incrementing loop with O_CREAT|O_EXCL.
				if(++(dpth->looped)>1)
				{
					logp("could not find any free data file entries out of the 15000^3 available!\n");
					return -1;
				}
			}
		}
	}
//printf("before incr_dpth: %s %04X/%04X/%04X\n", dpth->path, dpth->prim, dpth->seco, dpth->tert);
	mk_dpth(dpth);
//printf("after incr_dpth: %s\n", dpth->path);
	return 0;
}

static int init_dpth(struct dpth *dpth, const char *currentdata)
{
	char *tmp=NULL;
//logp("in init_dpth\n");
	dpth->looped=0;
	dpth->prim=0;
	dpth->seco=0;
	dpth->tert=0;
	
	if((dpth->prim=get_highest_entry(currentdata))<0)
	{
		// Could not open directory. Set all zeros.
		dpth->prim=0;
		mk_dpth(dpth);
		return 0;
	}
	mk_dpth_prim(dpth);
	if(!(tmp=prepend_s(currentdata, dpth->path, strlen(dpth->path))))
	{
		log_and_send("out of memory");
		return -1;
	}
	if((dpth->seco=get_highest_entry(tmp))<0)
	{
		// Could not open directory. Set zero.
		dpth->seco=0;
		mk_dpth(dpth);
		free(tmp);
		return 0;
	}
	free(tmp);
	mk_dpth_seco(dpth);
	if(!(tmp=prepend_s(currentdata, dpth->path, strlen(dpth->path))))
	{
		log_and_send("out of memory");
		return -1;
	}
	if((dpth->tert=get_highest_entry(tmp))<0)
	{
		// Could not open directory. Set zero.
		dpth->tert=0;
		mk_dpth(dpth);
		free(tmp);
		return 0;
	}
	// At this point, we have the latest data file. Increment to get the
	// next free one.
	incr_dpth(dpth);

//logp("init_dpth: %d/%d/%d\n", dpth->prim, dpth->seco, dpth->tert);
//logp("init_dpth: %s\n", dpth->path);

	return 0;
}

static int receive_unchanged_file(const char *fname, size_t flen, FILE *mp, const char *dpth, const char *statbuf, size_t slen)
{
	send_msg_fp(mp, 't', dpth, strlen(dpth));
	send_msg_fp(mp, 'r', statbuf, slen);
	send_msg_fp(mp, 'f', fname, flen);
	return 0;
}

static int do_rename(const char *oldpath, const char *newpath)
{
	if(rename(oldpath, newpath))
	{
		logp("could not rename '%s' to '%s': %s\n",
			newpath, oldpath, strerror(errno));
		return -1;
	}
	return 0;
}

static int receive_existing_file(const char *working, char cmd, const char *fname, size_t flen, const char *datadir, const char *currentdata, const char *newpath, FILE *mp, const char *dpth, const char *statbuf, size_t slen, struct stat *statp)
{
	int ret=0;
	char *opath=NULL;
	char *rpath=NULL;

	// already exists and has different mtime/ctime, send
	// signature
	logp("sig/delta file: (%s) %s\n", dpth, fname);

	if(!(opath=prepend_s(currentdata, dpth, strlen(dpth))))
	{
		async_write_str('e', "out of memory");
		ret=-1;
	}
	else if(build_path(datadir, dpth, strlen(dpth), &rpath))
	{
		log_and_send("build path failed");
		ret=-1;
	}
	else
	{
		FILE *fp=NULL;
		char delfor[64]="";
		char deltmp[64]="";
		char *delpath=NULL;
		char *delpathtmp=NULL;

		// Write to a temporary file first, then move it into place
		// when we know it is complete.
		// Otherwise, possible problems arise when picking up after a
		// power cut.
		snprintf(deltmp, sizeof(deltmp), "deltas.forward/%s.tmp", dpth);
		snprintf(delfor, sizeof(delfor), "deltas.forward/%s", dpth);
		if(!(delpath=prepend_s(working, delfor, strlen(delfor)))
		  || !(delpathtmp=prepend_s(working, deltmp, strlen(deltmp)))
		  || mkpath(&delpath)
		  || !(fp=open_file(delpathtmp, "wb"))
		  || transfer_file_in(NULL, fp, NULL))
			ret=-1;
		if(fp) fclose(fp);

		if(!ret)
		{
			// It is OK to put the complete file into place
			// before writing to the manifest - the file is
			// complete and should not cause any trouble.
			if(do_rename(delpathtmp, delpath))
			{
				ret=-1;
			}
			else
			{
				send_msg_fp(mp, 't', dpth, strlen(dpth));
				send_msg_fp(mp, 'r', statbuf, slen);
				send_msg_fp(mp, 'f', fname, flen);
				// Flush now so that we know what we have got
				// later if the power cuts out.
				fflush(mp);
			}
		}
		else unlink(delpathtmp);

		if(delpath) free(delpath);
		if(delpathtmp) free(delpathtmp);
	}
	if(opath) free(opath);
	if(rpath) free(rpath);
	return ret;
}

static int receive_new_file(char cmd, const char *fname, size_t flen, const char *datadir, const char *currentdata, FILE *mp, struct dpth *dpth, const char *statbuf, size_t slen, struct stat *statp)
{
	int ret=0;
	FILE *fp=NULL;
	char *rpath=NULL;
	char *datapth=NULL;

	if(build_path(datadir, dpth->path, strlen(dpth->path), &rpath))
	{
		log_and_send("build path failed");
		ret=-1;
	}
	else if(!(fp=open_file(rpath, "wb")))
	{
		log_and_send("make file failed");
		ret=-1;
	}
	else
	{
		// Do not bother writing to a tmpfile and then renaming in
		// the case of a full new file, as we do for receiving the
		// deltas. Problems cannot arise here in the same way is
		// in the delta case.
		logp("new file: (%s) %s\n", dpth->path, fname);
		ret=transfer_file_in(NULL, fp, NULL);
		close_fp(&fp);

		if(!ret)
		{
			send_msg_fp(mp, 't', dpth->path, strlen(dpth->path));
			send_msg_fp(mp, 'r', statbuf, slen);
			send_msg_fp(mp, 'f', fname, flen);
			// Flush now so that we know what we have got later if
			// the power cuts out.
			fflush(mp);
			incr_dpth(dpth);
		}
		else unlink(rpath);
	}
	if(rpath) free(rpath);
	if(datapth) free(datapth);
	return ret;
}

static int read_timestamp(const char *path, char buf[], size_t len)
{
	FILE *fp=NULL;
	char *cp=NULL;

	//if(!(fp=open_file(path, "rb")))
	// avoid alarming message
	if(!(fp=fopen(path, "rb")))
	{
		*buf=0;
		return -1;
	}
	fgets(buf, len, fp);
	fclose(fp);
	if((cp=strrchr(buf, '\n'))) *cp='\0';
	if(!*buf) return -1;
	return 0;
}

struct bu
{
	char *path;
	char *data;
	char *delta;
	char *timestamp;
	char *forward_timestamp;
	unsigned long index;
	unsigned long forward_index;
};

static int bu_cmp(const void *va, const void *vb)
{
	const struct bu *a=(struct bu *)va;
	const struct bu *b=(struct bu *)vb;
	if(     a->index > b->index) return 1;
	else if(a->index < b->index) return -1;
	return 0;
}

static void free_current_backups(struct bu **arr, int a)
{
	int b=0;
	for(b=0; b<a; b++)
	{
		if((*arr)[b].path)
			free((*arr)[b].path);
		if((*arr)[b].data)
			free((*arr)[b].data);
		if((*arr)[b].delta)
			free((*arr)[b].delta);
		if((*arr)[b].timestamp)
			free((*arr)[b].timestamp);
		if((*arr)[b].forward_timestamp)
			free((*arr)[b].forward_timestamp);
	}
	free(*arr);
}

static int get_current_backups(const char *basedir, struct bu **arr, int *a)
{
	int i=0;
	int j=0;
	int ret=0;
	DIR *d=NULL;
	char buf[32]="";
	struct dirent *dp=NULL;

	if(!(d=opendir(basedir)))
	{
		log_and_send("could not open backup directory");
		return -1;
	}
	while((dp=readdir(d)))
	{
		struct stat statp;
		char *fullpath=NULL;
		char *timestamp=NULL;
		char *forward=NULL;
		char *timestampstr=NULL;
		char *forward_timestampstr=NULL;

		if(dp->d_ino==0
		  || !strcmp(dp->d_name, ".")
		  || !strcmp(dp->d_name, ".."))
			continue;
		if(!(fullpath=prepend_s(basedir,
			dp->d_name, strlen(dp->d_name)))
		 || !(timestamp=prepend_s(fullpath,
			"timestamp", strlen("timestamp")))
		 || !(forward=prepend_s(fullpath,
			"forward", strlen("forward"))))
		{
			ret=-1;
			if(timestamp) free(timestamp);
			if(fullpath) free(fullpath);
			break;
		}
		if((!lstat(fullpath, &statp) && !S_ISDIR(statp.st_mode))
		  || lstat(timestamp, &statp) || !S_ISREG(statp.st_mode)
		  || read_timestamp(timestamp, buf, sizeof(buf))
		  || !(timestampstr=strdup(buf)))
		{
			free(fullpath);
			free(forward);
			free(timestamp);
			continue;
		}
		free(timestamp);

		if(!read_timestamp(forward, buf, sizeof(buf)))
			forward_timestampstr=strdup(buf);
		free(forward);

		if(!(*arr=(struct bu *)realloc(*arr,(i+1)*sizeof(struct bu)))
		  || !((*arr)[i].data=prepend_s(fullpath, "data", strlen("data")))
		  || !((*arr)[i].delta=prepend_s(fullpath, "deltas.reverse", strlen("deltas.reverse"))))
		{
			log_and_send("out of memory");
			free(timestampstr);
			free(fullpath);
			break;
		}
		(*arr)[i].path=fullpath;
		(*arr)[i].timestamp=timestampstr;
		(*arr)[i].index=strtoul(timestampstr, NULL, 10);
		if(forward_timestampstr)
		{
			(*arr)[i].forward_timestamp=forward_timestampstr;
			(*arr)[i].forward_index=strtoul(forward_timestampstr,
				NULL, 10);
		}
		else
		{
			(*arr)[i].forward_timestamp=NULL;
			(*arr)[i].forward_index=0;
		}
		i++;
	}
	closedir(d);

	if(*arr) qsort(*arr, i, sizeof(struct bu), bu_cmp);

	for(j=0; j<i-1; j++)
	{
		if(!(*arr)[j].forward_timestamp
		  || strcmp((*arr)[j].forward_timestamp, (*arr)[j+1].timestamp)
		  || (*arr)[j].forward_index!=(*arr)[j+1].index)
		{
			char msg[256]="";
			snprintf(msg, sizeof(msg), "%s does not match forward path from %s\n", (*arr)[j+1].path, (*arr)[j].path);
			log_and_send(msg);
			free_current_backups(arr, i);
			ret=-1;
			break;
		}
	}

	*a=i;
		  
	return ret;
}

static int get_new_timestamp(const char *basedir, char *buf, size_t s)
{
	int a=0;
	time_t t=0;
	const struct tm *ctm=NULL;
	unsigned long index=0;
	char tmpbuf[32]="";
	struct bu *arr=NULL;

	// Want to prefix the timestamp with an index that increases by
	// one each time. This makes it far more obvious which backup depends
	// on which - even if the system clock moved around. Take that,
	// bacula!

	// get_current_backups orders the array with the highest index number 
	// last
	if(get_current_backups(basedir, &arr, &a)) return -1;
	if(a) index=arr[a-1].index;

	free_current_backups(&arr, a);

	time(&t);
	ctm=localtime(&t);
	strftime(tmpbuf, sizeof(tmpbuf), "%Y-%m-%d %T", ctm);
	snprintf(buf, s, "%07lu %s", ++index, tmpbuf);

	return 0;
}

static int write_timestamp(const char *timestamp, const char *tstmp)
{
	FILE *fp=NULL;
	if(!(fp=open_file(timestamp, "wb"))) return -1;
	fprintf(fp, "%s\n", tstmp);
	fclose(fp);
	return 0;
}

static size_t block_len=RS_DEFAULT_BLOCK_LEN;
static size_t strong_len=RS_DEFAULT_STRONG_LEN;

static int make_rev_sig(const char *dst, const char *sig)
{
	gzFile dstp=NULL;
	FILE *sigp=NULL;
	rs_result result;
//logp("make rev sig: %s %s\n", dst, sig);

	if(!(dstp=gzopen_file(dst, "rb9"))
	  || !(sigp=open_file(sig, "wb")))
	{
		if(dstp) gzclose_fp(&dstp);
		return -1;
	}
	result=rs_sig_gzfile(dstp, sigp, block_len, strong_len, NULL);
	gzclose_fp(&dstp);
	fclose(sigp);
//logp("end of make rev sig\n");
	return result;
}

static int make_rev_delta(const char *src, const char *sig, const char *del)
{
	gzFile srcp=NULL;
	FILE *sigp=NULL;
	gzFile delp=NULL;
	rs_result result;
	rs_signature_t *sumset=NULL;

//logp("make rev delta: %s %s %s\n", src, sig, del);
	if(!(sigp=open_file(sig, "rb"))) return -1;
	if((result=rs_loadsig_file(sigp, &sumset, NULL))
	  || (result=rs_build_hash_table(sumset)))
	{
		fclose(sigp);
		rs_free_sumset(sumset);
		return result;
	}
	fclose(sigp);

//logp("make rev deltb: %s %s %s\n", src, sig, del);

	if(!(srcp=gzopen_file(src, "rb9"))
	  || !(delp=gzopen_file(del, "wb9")))
	{
		if(srcp) gzclose_fp(&srcp);
		rs_free_sumset(sumset);
		return -1;
	}

	result=rs_delta_gzfile(sumset, srcp, delp, NULL);
	rs_free_sumset(sumset);
	gzclose_fp(&srcp);
	gzclose_fp(&delp);

	return result;
}


static int gen_rev_delta(const char *sigpath, const char *deltadir, const char *oldpath, const char *finpath, const char *path)
{
	int ret=0;
	char *delpath=NULL;
	if(!(delpath=prepend_s(deltadir, path, strlen(path))))
	{
		logp("out of memory\n");
		ret=-1;
	}
	logp("Generating reverse delta...\n");
/*
	logp("delpath: %s\n", delpath);
	logp("finpath: %s\n", finpath);
	logp("sigpath: %s\n", sigpath);
	logp("oldpath: %s\n", oldpath);
*/
	if(mkpath(&delpath))
	{
		logp("could not mkpaths for: %s\n", delpath);
		ret=-1;
	}
	else if(make_rev_sig(finpath, sigpath))
	{
		logp("could not make signature from: %s\n", finpath);
		ret=-1;
	}
	else if(make_rev_delta(oldpath, sigpath, delpath))
	{
		logp("could not make delta from: %s\n", oldpath);
		ret=-1;
	}
	else unlink(sigpath);	
	if(delpath) free(delpath);
	return ret;
}

static int inflate_oldfile(const char *oldpath, const char *infpath)
{
	int ret=0;
	struct stat statp;
	FILE *source=NULL;
	FILE *dest=NULL;

	logp("inflating...\n");

	if(lstat(oldpath, &statp))
	{
		logp("could not lstat %s\n", oldpath);
		return -1;
	}

	if(!(dest=open_file(infpath, "wb")))
	{
		close_fp(&dest);
		return -1;
	}

	if(!statp.st_size)
	{
		// Empty file - cannot inflate.
		// just close the destination and we have duplicated a zero
		// length file.
		logp("asked to inflate zero length file: %s\n", oldpath);
		close_fp(&dest);
		return 0;
	}

	if(!(source=open_file(oldpath, "rb")))
	{
		close_fp(&dest);
		return -1;
	}

	if((ret=zlib_inflate(source, dest))!=Z_OK)
		logp("zlib_inflate returned: %d\n", ret);

	close_fp(&source);
	close_fp(&dest);
	return ret;
}

/* Need to make all the stuff that this does atomic so that existing backups
   never get broken, even if somebody turns the power off on the server. */ 
static int atomic_data_jiggle(const char *finishing, const char *working, const char *manifest, const char *current, const char *currentdata, const char *datadir, const char *datadirtmp, struct config *cconf)
{
	char cmd;
	int ret=0;
	size_t len=0;
	char *buf=NULL;
	char *newpath=NULL;
	char *oldpath=NULL;
	char *finpath=NULL;
	char *deltabdir=NULL;
	char *deltafdir=NULL;
	char *deltafpath=NULL;
	char *sigpath=NULL;
	struct stat statp;
	gzFile mp=NULL;

	logp("Doing the atomic data jiggle...\n");

	if(cconf->hardlinked_archive)
	  logp("Hardlinked archive is on - will not generate reverse deltas\n");
	else
	  logp("Hardlinked archive is off - will generate reverse deltas\n");

	if(!(mp=gzopen_file(manifest, "rb9"))) return -1;

	if(!(deltabdir=prepend_s(current,
		"deltas.reverse", strlen("deltas.reverse")))
	  || !(deltafdir=prepend_s(finishing,
		"deltas.forward", strlen("deltas.forward")))
	  || !(sigpath=prepend_s(current,
		"sig.tmp", strlen("sig.tmp"))))
	{
		logp("out of memory\n");
		gzclose_fp(&mp);
		return -1;
	}

	mkdir(datadir, 0777);
	while(!ret)
	{
		if(async_read_fp(NULL, mp, &cmd, &buf, &len))
		{
			break;
		}
		else
		{
			if(cmd!='t')
			{
				// ignore
				if(buf) { free(buf); buf=NULL; }
				continue;
			}
			if(buf[len]=='\n') buf[len]='\0';
			if(!(oldpath=prepend_s(currentdata,
				buf, strlen(buf)))
			  || !(newpath=prepend_s(datadirtmp,
				buf, strlen(buf)))
			  || !(finpath=prepend_s(datadir,
				buf, strlen(buf)))
			  || !(deltafpath=prepend_s(deltafdir,
				buf, strlen(buf))))
			{
				logp("out of memory\n");
				ret=-1;	
			}
			else if(!lstat(finpath, &statp)
			  && S_ISREG(statp.st_mode))
			{
				// Looks like an interrupted jiggle
				// did this file already.
				if(!lstat(deltafpath, &statp)
				  && S_ISREG(statp.st_mode))
				{
					logp("deleting unneeded forward delta: %s\n", deltafpath);
					unlink(deltafpath);
				}
				logp("skipping already present file: %s\n",
					finpath);
			}
			else if(mkpath(&finpath))
			{
				logp("could not create path for: %s\n",
					finpath);
				ret=-1;
			}
			else if(!lstat(deltafpath, &statp)
			  && S_ISREG(statp.st_mode))
			{
				char *cp=NULL;
				char *infpath=NULL;

				// Got a forward patch to do.
				// First, need to gunzip the old file,
				// otherwise the librsync patch will take
				// forever, because it will be doing seeks
				// all over the place, and gzseeks are slow.

				if(!(infpath=strdup(deltafpath)))
				{
					logp("out of memory\n");
					ret=-1;
				}
				else if(!(cp=strrchr(infpath, '.')))
				{
					logp("could not strip the suffix from '%s'\n", infpath);
					ret=-1;
				}
				*cp='\0';

				logp("Fixing up: %s\n", buf);
				if(!ret && inflate_oldfile(oldpath, infpath))
				{
					logp("error when inflating old file: %s\n", oldpath);
					ret=-1;
				}

				// Gzip the result with TRUE;
				if(!ret && do_patch(infpath, deltafpath,
					newpath, TRUE))
				{
					logp("error when patching\n");
					ret=-1;
					// Remove anything that got written.
					unlink(newpath);
				}

				// Get rid of the inflated old file.
				// This will also remove it if there was an
				// error.
				unlink(infpath);
				free(infpath);

				if(!ret)
				{
					// Need to generate a reverse diff,
					// unless we are keeping a hardlinked
					// archive.
					if(!cconf->hardlinked_archive)
					{
					  if(gen_rev_delta(sigpath, deltabdir,
					    oldpath, newpath, buf))
						ret=-1;
					}

					// Power interruptions should be
					// recoverable. If it happens before
					// this point, the data jiggle for
					// this file has to be done again.
					// Once finpath is in place, no more
					// jiggle is required.

					// Use the fresh new file.
					if(!ret && do_rename(newpath, finpath))
						ret=-1;
					else if(!ret)
					{
					  // Remove the forward delta, as it is
					  // no longer needed. There is a
					  // reverse diff and the finished
					  // finished file is in place.
					  logp("Deleting delta.forward...\n");
					  unlink(deltafpath);

					  // Remove the old file. If a power
					  // cut happens just before this,
					  // the old file will hang around
					  // forever.
					  // TODO: Put in something to
					  // detect this.
					  // ie, both a reverse delta and the
					  // old file exist.
					  if(!cconf->hardlinked_archive)
					  {
					    logp("Deleting oldpath...\n");
					    unlink(oldpath);
					  }
					}
				}
			}
			else if(!lstat(newpath, &statp)
			     && S_ISREG(statp.st_mode))
			{
				// Use the fresh new file.
				// This needs to happen after checking
				// for the forward delta, because the
				// patching stuff writes to newpath.
				logp("Using newly received file\n");
				if(do_rename(newpath, finpath)) ret=-1;
			}
			else if(!lstat(oldpath, &statp)
			  && S_ISREG(statp.st_mode))
			{
				// Use the old unchanged file.
				// Hard link it first.
				logp("Hard linking to old file: %s\n", buf);
				if(link(oldpath, finpath))
				{
					logp("could not hard link '%s' to '%s': %s\n", finpath, oldpath, strerror(errno));
					ret=-1;
				}
				else
				{
					// If we are not keeping a hardlinked
					// archive, delete the old link.
					if(!cconf->hardlinked_archive)
					{
						logp("Unlinking old file: %s\n",
							 oldpath);
						unlink(oldpath);
					}
				}
			}
			else
			{
				logp("could not find: %s\n", oldpath);
				ret=-1;
			}

			if(oldpath) { free(oldpath); oldpath=NULL; }
			if(newpath) { free(newpath); newpath=NULL; }
			if(finpath) { free(finpath); finpath=NULL; }
			if(deltafpath) { free(deltafpath); deltafpath=NULL; }
		}
		if(buf) { free(buf); buf=NULL; }
	}
	gzclose_fp(&mp);

	if(!ret)
	{
		// Remove the temporary data directory, we have now removed
		// everything useful from it.
		recursive_delete(deltafdir, NULL, FALSE /* do not del files */);
	}
	if(deltabdir) free(deltabdir);
	if(deltafdir) free(deltafdir);
	if(sigpath) free(sigpath);
	if(buf) { free(buf); buf=NULL; }
	return ret;
}

static int compress_file(const char *src, const char *dst)
{
	int res;
	int got;
	FILE *mp=NULL;
	gzFile zp=NULL;
	char buf[16000];

	if(!(mp=open_file(src, "rb"))
	  || !(zp=gzopen_file(dst, "wb9")))
	{
		close_fp(&mp);
		gzclose_fp(&zp);
		return -1;
	}
	while((got=fread(buf, 1, sizeof(buf), mp))>0)
	{
		res=gzwrite(zp, buf, got);
		if(res!=got)
		{
			logp("compressing manifest - read %d but wrote %d\n",
				got, res);
			close_fp(&mp);
			gzclose_fp(&zp);
			return -1;
		}
	}
	close_fp(&mp);
	gzclose_fp(&zp);
	return 0;
}

static int compress_log(const char *current)
{
	char *logpath=NULL;
	char *logpathtmp=NULL;

	if(!(logpath=prepend_s(current, "log.gz", strlen("log.gz")))
	  || !(logpathtmp=prepend_s(current, "log", strlen("log"))))
		return -1;
	
	// Need to compress the log.
	logp("Compressing log...\n");
	if(compress_file(logpathtmp, logpath))
	{
		unlink(logpath);
		free(logpath);
		free(logpathtmp);
		return -1;
	}
	// succeeded - get rid of the uncompressed version
	unlink(logpathtmp);
	free(logpath);
	free(logpathtmp);
	return 0;
}

int finish_backup(const char *basedir, const char *working, const char *current, const char *currentdata, const char *finishing, struct config *cconf)
{
	int ret=0;
	struct stat statp;
	char *manifest=NULL;
	char *manifesttmp=NULL;
	char *datadir=NULL;
	char *datadirtmp=NULL;
	char *currentdup=NULL;
	char *currentduptmp=NULL;
	char *currentdupdata=NULL;
	char *forward=NULL;
	char *timestamp=NULL;
	char *fullrealcurrent=NULL;
	char *deleteme=NULL;
	char *logpath=NULL;
	int len=0;
	char realcurrent[256]="";
	FILE *logfp=NULL;

	if((len=readlink(current, realcurrent, sizeof(realcurrent)-1))<0)
		len=0;
	realcurrent[len]='\0';

	if(!(datadir=prepend_s(finishing, "data", strlen("data")))
	  || !(datadirtmp=prepend_s(finishing, "data.tmp", strlen("data.tmp")))
	  || !(manifesttmp=prepend_s(finishing, "manifest", strlen("manifest")))
	  || !(manifest=prepend_s(finishing, "manifest.gz", strlen("manifest.gz")))
	  || !(currentdup=prepend_s(finishing, "currentdup", strlen("currentdup")))
	  || !(currentduptmp=prepend_s(finishing, "currentdup.tmp", strlen("currentdup.tmp")))
	  || !(currentdupdata=prepend_s(currentdup, "data", strlen("data")))
	  || !(forward=prepend_s(currentdup, "forward", strlen("forward")))
	  || !(timestamp=prepend_s(finishing, "timestamp", strlen("timestamp")))
	  || !(fullrealcurrent=prepend_s(basedir, realcurrent, strlen(realcurrent)))
	  || !(deleteme=prepend_s(basedir, "deleteme", strlen("deleteme")))
	  || !(logpath=prepend_s(finishing, "log", strlen("log"))))
	{
		ret=-1;
		goto endfunc;
	}

	if(!(logfp=open_file(logpath, "ab")) || set_logfp(logfp))
	{
		ret=-1;
		goto endfunc;
	}

	if(!lstat(manifesttmp, &statp) && lstat(manifest, &statp))
	{
		// Need to compress the manifest.
		logp("Compressing manifest...\n");
		if(compress_file(manifesttmp, manifest))
		{
			ret=-1;
			unlink(manifest);
			goto endfunc;
		}
		// succeeded - get rid of the uncompressed version
		unlink(manifesttmp);
	}

	if(!lstat(current, &statp)) // Had a previous backup
	{
		FILE *fwd=NULL;
		char tstmp[64]="";

		if(lstat(currentdup, &statp))
		{
			// Have not duplicated the current backup yet.
			if(!lstat(currentduptmp, &statp))
			{
				logp("Removing previous currentduptmp directory: %s\n", currentduptmp);
				if(recursive_delete(currentduptmp,
					NULL, TRUE /* del files */))
				{
					logp("Could not delete %s\n",
						currentduptmp);
					ret=-1;
					goto endfunc;
				}
			}
			if(recursive_hardlink(current, currentduptmp)
			  || do_rename(currentduptmp, currentdup))
			{
				ret=-1;
				goto endfunc;
			}
		}

		if(read_timestamp(timestamp, tstmp, sizeof(tstmp)))
		{
			logp("could not read timestamp file: %s\n", timestamp);
			ret=-1;
			goto endfunc;
		}

		// Put forward reference in, indicating the timestamp of
		// the working directory (which will soon become the current
		// directory).
		if(!(fwd=open_file(forward, "wb")))
		{
			log_and_send("could not create forward file");
			ret=-1;
			goto endfunc;
		}
		fprintf(fwd, "%s\n", tstmp);
		close_fp(&fwd);

		if(atomic_data_jiggle(finishing,
			working, manifest, currentdup,
			currentdupdata,
			datadir, datadirtmp, cconf))
		{
			logp("could not finish up backup.\n");
			ret=-1;
			goto endfunc;
		}

		// Remove the temporary data directory, we have now removed
		// everything useful from it.
		recursive_delete(datadirtmp, NULL, TRUE /* del files */);

		// Clean up the currentdata directory - this is now the 'old'
		// currentdata directory. Any files that were deleted from
		// the client will be left in there, so call recursive_delete
		// with the option that makes it not delete files.
		// This will have the effect of getting rid of unnecessary
		// directories.
		recursive_delete(currentdupdata, NULL, FALSE /* do not del files */);

		// Rename the old current to something that we know to
		// delete.
		if(do_rename(fullrealcurrent, deleteme))
		{
			ret=-1;
			goto endfunc;
		}
	}
	else
	{
		// No previous backup, just put datadirtmp in the right place.
		if(do_rename(datadirtmp, datadir))
		{
			ret=-1;
			goto endfunc;
		}
	}

	if(!lstat(deleteme, &statp))
	{
		// Rename the currentdup directory...
		// IMPORTANT TODO: read the path to fullrealcurrent
		// from the deleteme timestamp.
		do_rename(currentdup, fullrealcurrent);

		recursive_delete(deleteme, NULL, TRUE /* delete all */);
	}

	// Rename the finishing symlink so that it becomes the current symlink
	do_rename(finishing, current);

	end_filecounter_backup(0);
	logp("Backup completed.\n");
	set_logfp(NULL); // will close logfp.
	compress_log(current);

endfunc:
	if(datadir) free(datadir);
	if(datadirtmp) free(datadirtmp);
	if(manifest) free(manifest);
	if(manifesttmp) free(manifesttmp);
	if(currentdup) free(currentdup);
	if(currentduptmp) free(currentduptmp);
	if(currentdupdata) free(currentdupdata);
	if(forward) free(forward);
	if(timestamp) free(timestamp);
	if(fullrealcurrent) free(fullrealcurrent);
	if(deleteme) free(deleteme);
	if(logpath) free(logpath);

	return ret;
}

int remove_old_backups(const char *basedir, int keep)
{
	int a=0;
	int b=0;
	int ret=0;
	int del=0;
	struct bu *arr=NULL;

	logp("in remove_old_backups\n");

	if(get_current_backups(basedir, &arr, &a)) return -1;

	// Find the cut-off point.
	del=a-keep;

	// Trim from the back so that we do not end up with any
	// directories with broken dependencies.
	for(b=0; b<a; b++)
	{
		if(!ret && b<del)
		{
			logp("keeping %d backups\n", keep);
			logp("deleting %s\n", arr[b].timestamp);
			ret=recursive_delete(arr[b].path, NULL, TRUE);
		}
	}

	free_current_backups(&arr, a);

	return ret;
}

static void free_manbuf(char **buf)
{
	if(buf && *buf) { free(*buf); *buf=NULL; }
}

static void free_manbufs(char **fulldpth, char **sbuf, char **pbuf, char **dbuf)
{
	//logp("in free_manbufs\n");
	free_manbuf(fulldpth);
	free_manbuf(sbuf);
	free_manbuf(pbuf);
	free_manbuf(dbuf);
}

// Continue through the manifest and find the next entry for which we may have
// to send a signature to the client.
static int read_next_manifest_sig_line(FILE *fp, gzFile zp, const char *currentdata,
	char **fulldpth,
	char **pbuf, char **sbuf, char **dbuf,
	size_t *plen, size_t *slen, size_t *dlen,
	const char *sigcurpath, char **clientlatest)
{
	char xcmd='\0';
	size_t xlen=0;
	char *xbuf=NULL;

	//logp("in read_next: %s\n", *clientlatest);
	// If we find a path on the server, we want to generate and send
	// its signature.
	while(!async_read_fp(fp, zp, &xcmd, &xbuf, &xlen))
	{
		if(xbuf[xlen]=='\n') xbuf[xlen]='\0';
		if(xcmd=='t')
		{
			free_manbufs(fulldpth, sbuf, pbuf, dbuf);
			*dbuf=xbuf;
			*dlen=xlen;
			continue;
		}
		if(*dbuf)
		{
			if(!*sbuf)
			{
				if(xcmd=='r')
				{
					*sbuf=xbuf;
					*slen=xlen;
					xbuf=NULL;
					continue;
				}
				else
				{
					logp("unexpected line in manifest after dbuf line: %c %s\n", xcmd, xbuf);
					free_manbuf(&xbuf);
					return -1;
				}
			}
			else if(xcmd=='d' || xcmd=='f' || xcmd=='F')
			{
				free_manbuf(pbuf);
				*pbuf=xbuf;
				*plen=xlen;
				xbuf=NULL;
				if((sigcurpath && pathcmp(sigcurpath, *pbuf)>0)
				  || (clientlatest && *clientlatest && pathcmp(*clientlatest, *pbuf)>0))
				{
					//logp("read_next_sig_lines: %s\n",
					//	*pbuf);
					//logp("read_next_sig_linec: %s\n",
					//	sigcurpath);
					logp("client is ahead, skipping %s\n",
						*pbuf);
					if(*clientlatest) logp("clientlatest: %s\n", *clientlatest);
					free_manbufs(fulldpth,
						sbuf, pbuf, dbuf);
					continue;
				}
				if(clientlatest && *clientlatest && pathcmp(*clientlatest, *pbuf)<=0)
				{
					logp("caught up with client: %s\n", *pbuf);
					free(*clientlatest); *clientlatest=NULL;
				}
				break;
			}
		}

		free_manbuf(&xbuf);
	}
	if(fulldpth && *dbuf
	  && !(*fulldpth=prepend_s(currentdata, *dbuf, strlen(*dbuf))))
		return -1;
	return 0;
}

static int do_receive(const char *working,
	const char *currentdata, const char *datadirtmp,
	const char *tstmp, FILE **mp, const char *basedir, int keep,
	const char *newpath, struct dpth *dpth,
	const char *sigcurpath, // where we are with sending the sig
	int *sendendofsig,
	char **clientlatest
	)
{
	int ret=1;
	size_t wlen=0;

	static char pcmd='\0';
	static size_t slen=0;
	static size_t dlen=0;
	static size_t plen=0;
	static char *pbuf=NULL;
	static char *statbuf=NULL;
	static char *datapth=NULL;
	static struct stat statp;

//logp("do_receive begin\n");

	if(pcmd=='\0')
	{
	 size_t rlen=0;
	 char rcmd='\0';
	 char *rbuf=NULL;
	 if(async_rw(&rcmd, &rbuf, &rlen, '\0', NULL, &wlen)) return -1;
	 if(rbuf)
	 {
//logp("got: %c:%s\n", rcmd, rbuf);
	    switch(rcmd)
	    {
		case 't':
			// Given a path on the server for sig/sync stuff.
			// stat and filename should follow.
			if(datapth) free(datapth);
			datapth=rbuf;
			dlen=rlen;
			rbuf=NULL;
			break;
		case 'r':
			// Given the file stats.
			decode_stat(rbuf, &statp);
			if(statbuf) free(statbuf);
			statbuf=rbuf;
			slen=rlen;
			rbuf=NULL;
			break;
		case 'c':
			if(!strcmp(rbuf, "backupend"))
			{
				ret=0;
			}
			else
			{
				logp("do_receive: unexpected cmd: %c:%s\n",
					rcmd, rbuf);
				ret=-1;
			}
			if(rbuf) free(pbuf);
			if(pbuf) { free(pbuf); pbuf=NULL; } plen=0;
			if(statbuf) { free(statbuf); statbuf=NULL; } slen=0;
			if(datapth) { free(datapth); datapth=NULL; } dlen=0;
			return ret;
		case 'w':
			logp("WARNING: %s\n", rbuf);
			do_filecounter(rcmd, 0);
			if(rbuf) free(pbuf);
			break;
		default:
			// should now have a set - pbuf will
			// be the path on the client (unless it is an error -
			// this will get checked later).
			if(pbuf) free(pbuf);
			pcmd=rcmd;
			pbuf=rbuf;
			plen=rlen;
			rbuf=NULL;
			break;
	    }
	 }
	 if(rbuf) { free(rbuf); rbuf=NULL; }
	}

	if(ret==1 && sigcurpath && !*sendendofsig && pbuf && statbuf)
	{
		if(pcmd=='f' // Unchanged file
		  || (pcmd=='F' && !datapth)) // New file
		{
			int pcmp;
			if((pcmp=pathcmp(sigcurpath, pbuf))<=0)
			{
				// Send end of sig if client is ahead or level.
				//logp("pcmp: %d\n", pcmp);
				//logp("pathcmps: '%s'\n", sigcurpath);
				//logp("pathcmpc: '%s'\n", pbuf);
				//logp("send end of sig in do_receive\n");
				(*sendendofsig)++;
				return 1;
			}
		}
		else if(pcmd=='F' && datapth)
		{
			// Changed file, but have not finished sending the sig
			if(pathcmp(sigcurpath, pbuf)==0)
			{
				//logp("continue send sig in do_receive\n");
				// Try to write a little.
	 			if(async_rw(NULL, NULL, '\0',
					'\0', NULL, &wlen)) return -1;
				return 1;
			}
		}
	}

	if(ret==1 && pbuf && statbuf)
	{
	  switch(pcmd)
	  {
		// Will have a set of pbuf/statbuf/datapth by this point.
		// datapth can be NULL.
		case 'f':
logp("unchanged file: (%s) %s\n", datapth, pbuf);
			receive_unchanged_file(pbuf, plen, *mp,
				datapth, statbuf, slen);
			do_filecounter(pcmd, 0);
			break;
		case 'F': 
			if(datapth)
			{
				if(receive_existing_file(working, pcmd, pbuf,
					plen, datadirtmp, currentdata,
					newpath, *mp, datapth,
					statbuf, slen, &statp)) ret=-1;
				else do_filecounter('x', 0);
			}
			else if(receive_new_file(pcmd, pbuf, plen,
				datadirtmp, currentdata, *mp, dpth,
				statbuf, slen, &statp)) ret=-1;
				else do_filecounter(pcmd, 0);
			break;
		case 'd': 
			if(receive_dir(pcmd, pbuf, plen, *mp,
				statbuf, slen)) ret=-1;
			else do_filecounter(pcmd, 0);
			break;
		case 'l': 
		case 'L':
			if(receive_link(pcmd, pbuf, plen, *mp,
				statbuf, slen)) ret=-1;
			else do_filecounter(pcmd, 0);
			break;
		case 's':
			if(receive_special(pcmd, pbuf, plen, *mp,
				statbuf, slen)) ret=-1;
			else do_filecounter(pcmd, 0);
			break;
		default:
			logp("unknown cmd: %c\n", pcmd);
			ret=-1;
			break;
	  }

	  // Set clientlatest so that we can skip over signatures that
	  // the client does not need.
	  if(*clientlatest) free(*clientlatest);
	  if(pbuf) { *clientlatest=pbuf; } pbuf=NULL; plen=0;
	  if(statbuf) { free(statbuf); statbuf=NULL; } slen=0;
	  if(datapth) { free(datapth); datapth=NULL; } dlen=0;
	  pcmd='\0';
	}
//logp("do receive returning: %d\n", ret);
	return ret;
}

static int start_sigjob(gzFile sigfp, rs_job_t **sigjob, rs_filebuf_t **infb, rs_filebuf_t **outfb, rs_buffers_t *rsbuf)
{
	memset(rsbuf, 0, sizeof(*rsbuf));
	if(!(*sigjob=rs_sig_begin(block_len, strong_len)))
	{
		logp("could not start signature job.\n");
		return -1;
	}
	if(!(*infb=rs_filebuf_new(NULL, NULL, sigfp, -1, ZTYPE_NONE, rs_inbuflen)))
	{
		logp("could not rs_filebuf_new for infb.\n");
		return -1;
	}
	if(!(*outfb=rs_filebuf_new(NULL, NULL, NULL, async_get_fd(), ZTYPE_DEFLATE, rs_outbuflen)))
	{
		logp("could not rs_filebuf_new for in_outfb.\n");
		return -1;
	}

	return 0;
}

static void free_sigjob(rs_job_t **sigjob, rs_filebuf_t **infb, rs_filebuf_t **outfb, gzFile *sigfp, int *sendendofsig, char **sigcurpath)
{
	//logp("in free_sigjob\n");
	if(*sigjob) { rs_job_free(*sigjob); *sigjob=NULL; }
	if(*outfb) { rs_filebuf_free(*outfb); *outfb=NULL; }
	if(*infb) { rs_filebuf_free(*infb); *infb=NULL; }
	if(*sigfp) { gzclose_fp(sigfp); }
	if(*sigcurpath) { free(*sigcurpath); *sigcurpath=NULL; }
	*sendendofsig=0;
}

static int sigjob_stuff(gzFile *fp, gzFile *sigfp, rs_job_t **sigjob, rs_filebuf_t **infb, rs_filebuf_t **outfb, rs_buffers_t *rsbuf, const char *currentdata, char **pbuf, char **sbuf, char **dbuf, size_t *plen, size_t *slen, size_t *dlen, char **fulldpth, int *sendendofsig, char **sigcurpath, char **clientlatest)
{
	if(!*sigjob)
	{
		if(!*pbuf && !*sbuf && !*dbuf)
		{
			if(read_next_manifest_sig_line(NULL, *fp, currentdata,
			  fulldpth, pbuf, sbuf, dbuf, plen, slen, dlen,
			  *sigcurpath, clientlatest))
			{
				free_sigjob(sigjob,
					infb, outfb, sigfp, sendendofsig,
					sigcurpath);
				return -1;
			}

			if(!*dbuf || !*sbuf || !*pbuf)
			{
				*fp=NULL;
				logp("okmanifestend\n");
				// Have no manifest left - send an 'end of
				// manifest'  message.
				if(async_write('c',
				  "okmanifestend", strlen("okmanifestend")))
					return -1;
			}
			else if(fulldpth && *fulldpth)
			{
				struct stat statp;
				// Make sure the file exists and is non-zero
				// length.
				if(lstat(*fulldpth, &statp)
				  || !statp.st_size)
				{
				  logp("skipping sig for %s\n", *fulldpth);
				  free_manbufs(fulldpth, sbuf, pbuf, dbuf);
				}
			}
		}

		if(*pbuf || *sbuf || *dbuf)
		{
			if(plen && async_append_all_to_write_buffer('p',
				*pbuf, plen)) return 0;
			if(slen && async_append_all_to_write_buffer('r',
				*sbuf, slen)) return 0;
			if(dlen && async_append_all_to_write_buffer('t',
				*dbuf, dlen)) return 0;

			// start sending the signature straight away.
			logp("sending sig for %s\n", *dbuf);
			logp("(%s)\n", *pbuf);
			if(!(*sigfp=gzopen_file(*fulldpth, "rb9"))
			  || start_sigjob(*sigfp,
				sigjob, infb, outfb, rsbuf))
			{
				free_sigjob(sigjob, infb, outfb, sigfp,
				  sendendofsig, sigcurpath);
				free_manbufs(fulldpth, sbuf, pbuf, dbuf);
				return -1;
			}
			if(*sigcurpath) free(*sigcurpath);
			*sigcurpath=strdup(*pbuf);
			free_manbufs(fulldpth, sbuf, pbuf, dbuf);
		}
	}
	else
	{
		if(!*sendendofsig)
		{
			rs_result sigresult;

			sigresult=rs_async(*sigjob, rsbuf, *infb, *outfb);
			//logp("after sending sig: %d (%s)\n", sigresult, *sigcurpath);
			if(sigresult==RS_DONE)
			{
				(*sendendofsig)++;
				//logp("sendendofsig for: %s\n", *sigcurpath);
				free_manbufs(fulldpth, sbuf, pbuf, dbuf);
			}
			else if(sigresult==RS_BLOCKED || sigresult==RS_RUNNING)
			{
				//logp("RS_BLOCKED/RUNNING\n");
				// keep going round the loop.
			}
			else
			{
				logp("error in rs_async: %d\n", sigresult);
				free_sigjob(sigjob, infb, outfb, sigfp,
				  sendendofsig, sigcurpath);
				free_manbufs(fulldpth, sbuf, pbuf, dbuf);
				return -1;
			}
		}
	}
	return 0;
}

static void reset_timers(void)
{
	settimers(0, 100);
}

static int backup_select(gzFile cmanfp, const char *working, const char *currentdata, const char *datadirtmp, const char *tstmp, FILE **mp, const char *basedir, int keep, const char *newpath, struct dpth *dpth, const char *startdir)
{
	int ret;
	gzFile fp=cmanfp;
	gzFile sigfp=NULL;
	rs_filebuf_t *infb=NULL;
	rs_filebuf_t *outfb=NULL;
	rs_job_t *sigjob=NULL;
	rs_buffers_t rsbuf;

	size_t plen=0;
	size_t slen=0;
	size_t dlen=0;
	char *pbuf=NULL; // path on client
	char *sbuf=NULL; // stat line
	char *dbuf=NULL; // path on server
	char *fulldpth=NULL; // full path on server

	char *sigcurpath=NULL;
	int sendendofsig=0;

	char *clientlatest=strdup(startdir);

	reset_timers();

	// Have no manifest - send an 'end of manifest'  message.
	if(!fp)
	{
		logp("have no manifest - send okmanifestend\n");
		if(async_write('c', "okmanifestend", strlen("okmanifestend")))
		{
			logp("error when sending okmanifestend\n");
			if(clientlatest) free(clientlatest);
			return -1;
		}
	}

	while(1)
	{
		if(fp)
		{
			//logp("in sigjob_stuff - sigjob? %s sendend: %d cur: %s clientlatest: %s\n", sigjob?"yes":"no", sendendofsig, sigcurpath, clientlatest);
			if(sigjob_stuff(&fp, &sigfp, &sigjob, &infb, &outfb,
			  &rsbuf, currentdata,
			  &pbuf, &sbuf, &dbuf,
			  &plen, &slen, &dlen, &fulldpth,
			  &sendendofsig, &sigcurpath, &clientlatest))
			{
				logp("error in sigjob_stuff\n");
				if(clientlatest) free(clientlatest);
				return -1;
			}
		}

		if(sendendofsig)
		{
			size_t wlen;
			const char *endfile="endfile";
//logp("endsig\n");
			wlen=strlen(endfile);

			if(!async_append_all_to_write_buffer('x',
				endfile, &wlen))
			{
				if(!wlen)
				{
					// Managed to write it.
					free_sigjob(&sigjob, &infb,
					  &outfb, &sigfp,
					  &sendendofsig, &sigcurpath);
				}
			}
//logp("after endsig\n");
			// No error if it was not written, though.
			// Keep calm and carry on.
		}
		ret=do_receive(working, currentdata, datadirtmp,
			tstmp, mp, basedir, keep,
			newpath, dpth,
			sigcurpath, &sendendofsig, &clientlatest);
		if(ret<0)
		{
			logp("error in do_receive\n");
			free_sigjob(&sigjob, &infb, &outfb, &sigfp,
				&sendendofsig, &sigcurpath);
			free_manbufs(&fulldpth, &sbuf, &pbuf, &dbuf);
			if(clientlatest) free(clientlatest);
			return -1;
		}
		else if(!ret)
		{
			free_sigjob(&sigjob, &infb, &outfb, &sigfp,
				&sendendofsig, &sigcurpath);
			free_manbufs(&fulldpth, &sbuf, &pbuf, &dbuf);
			if(clientlatest) free(clientlatest);
			return 0; // backupend
		}
	}
	free_sigjob(&sigjob, &infb, &outfb, &sigfp,
		&sendendofsig, &sigcurpath);
	if(clientlatest) free(clientlatest);
	return 0;
}

static int do_backup(const char *basedir, const char *current, const char *working, const char *currentdata, const char *finishing, const char *startdir, struct config *cconf, const char *manifesttmp, const char *forward)
{
	int ret=0;
	char msg[256]="";
	FILE *mp=NULL;
	// The timestamp file of this backup
	char *timestamp=NULL;
	// Where the new file generated from the delta temporarily goes
	char *newpath=NULL;
	// path to the last manifest
	char *cmanifest=NULL;
	// Real path to the working directory
	char *realworking=NULL;
	char tstmp[64]="";
	char *datadirtmp=NULL;
	// Where to write messages
	char *logpath=NULL;
	FILE *logfp=NULL;

	struct dpth dpth;

	gzFile cmanfp=NULL;

	logp("in do_backup\n");

	if(!(timestamp=prepend_s(working, "timestamp", strlen("timestamp")))
	  || !(newpath=prepend_s(working, "patched.tmp", strlen("patched.tmp")))
	  || !(cmanifest=prepend_s(current, "manifest.gz", strlen("manifest.gz")))
	  || !(datadirtmp=prepend_s(working, "data.tmp", strlen("data.tmp")))
	  || get_new_timestamp(basedir, tstmp, sizeof(tstmp))
	  || !(realworking=prepend_s(basedir, tstmp, strlen(tstmp)))
	  || !(logpath=prepend_s(realworking, "log", strlen("log"))))
	{
		log_and_send("out of memory");
		ret=-1;
	}
	else if(mkdir(realworking, 0777))
	{
		snprintf(msg, sizeof(msg),
		  "could not mkdir for increment: %s", working);
		log_and_send(msg);
		ret=-1;
	}
	else if(!(logfp=open_file(logpath, "ab")) || set_logfp(logfp))
	{
		snprintf(msg, sizeof(msg),
		  "could not open log file: %s", logpath);
		log_and_send(msg);
		ret=-1;
	}
	else if(symlink(tstmp, working)) // relative link to the real work dir
	{
		snprintf(msg, sizeof(msg),
		  "could not point working symlink to: %s", realworking);
		ret=-1;
	}
	else if(mkdir(datadirtmp, 0777))
	{
		snprintf(msg, sizeof(msg),
		  "could not mkdir for datadir: %s", datadirtmp);
		log_and_send(msg);
		ret=-1;
	}
	else if(write_timestamp(timestamp, tstmp))
	{
		snprintf(msg, sizeof(msg),
		  "unable to write timestamp %s", timestamp);
		log_and_send(msg);
		ret=-1;
	}
	else if(!(mp=open_file(manifesttmp, "wb")))
	{
		snprintf(msg, sizeof(msg),
		  "could not open %s", manifesttmp);
		async_write_str('e', msg);
		ret=-1;
	}
	if(!ret && init_dpth(&dpth, currentdata))
	{
		log_and_send("could not init_dpth\n");
		ret=-1;
	}

	if(!ret && !(cmanfp=gzopen_file(cmanifest, "rb9")))
	{
		// only an error if we have an old manifest
		struct stat statp;
		if(!lstat(cmanifest, &statp))
		{
			logp("could not open old manifest %s\n", cmanifest);
			ret=-1;
		}
	}

	if(cmanfp) logp("sending: %s\n", cmanifest);
	if(!ret && backup_select(cmanfp, working, currentdata, datadirtmp,
		tstmp, &mp, basedir, cconf->keep, newpath, &dpth,
		startdir))
	{
		log_and_send("backup_select failed");
		ret=-1;
	}

	if(!ret)
	{
		// will not write anything more to
		// the new manifest
		// finish_backup will open it again
		// for reading
		close_fp(&mp);

		async_write_str('c', "okbackupend");
		logp("Backup ending - disconnect from client.\n");

		// Close the connection with the client, the rest of the job
		// we can do by ourselves.
		async_free();

		// Move the symlink to indicate that we are now in the end
		// phase. 
		if(do_rename(working, finishing))
			ret=-1;
		else
		{
			set_logfp(NULL); // does an fclose on logfp.
			// finish_backup will open logfp again
			ret=finish_backup(basedir,
				working, current, currentdata,
				finishing, cconf);
			if(!ret && cconf->keep>0)
				ret=remove_old_backups(basedir, cconf->keep);
		}
	}
	gzclose_fp(&cmanfp);
	close_fp(&mp);
	if(timestamp) free(timestamp);
	if(newpath) free(newpath);
	if(cmanifest) free(cmanifest);
	if(datadirtmp) free(datadirtmp);
	if(logpath) free(logpath);
	set_logfp(NULL); // does an fclose on logfp.
	return ret;
}

static int check_regex(regex_t *regex, const char *buf)
{
	if(!regex) return 1;
	switch(regexec(regex, buf, 0, NULL, 0))
	{
		case 0: return 1;
		case REG_NOMATCH: return 0;
		default: return 0;
	}
}

static int list_manifest(const char *fullpath, regex_t *regex)
{
	int ret=0;
	int quit=0;
	gzFile fp=NULL;
	char *manifest=NULL;

	char scmd;
	char *sbuf=NULL;
	size_t slen=0;
	char pcmd;
	char *pbuf=NULL;
	size_t plen=0;

	char lcmd;
	char *lbuf=NULL;
	size_t llen=0;

	if(!(manifest=prepend_s(fullpath, "manifest.gz", strlen("manifest.gz"))))
	{
		log_and_send("out of memory");
		return -1;
	}
	if(!(fp=gzopen_file(manifest, "rb9")))
	{
		log_and_send("could not open manifest");
		free(manifest);
		return -1;
	}
	free(manifest);

	while(!quit)
	{
		//logp("list manifest loop\n");
		// Need to parse while sending, to take note of the regex.
		if(async_read_fp(NULL, fp, &scmd, &sbuf, &slen))
			quit++;
		else if(scmd=='t')
		{
			if(sbuf) { free(sbuf); sbuf=NULL; }
			continue;
		}
		else if(scmd!='r')
		{
			log_and_send("unexpected scmd in manifest\n");
			quit++; ret=-1;
		}
		else if(async_read_fp(NULL, fp, &pcmd, &pbuf, &plen))
		{
			log_and_send("missing cmd in manifest\n");
			quit++; ret=-1;
		}
		else if(pcmd!='f' && pcmd!='d' && pcmd!='l' && pcmd!='L')
		{
			log_and_send("unexpected pcmd in manifest\n");
			quit++; ret=-1;
		}
		// Links have a second line.
		else if((pcmd=='l' || pcmd=='L') && 
			async_read_fp(NULL, fp, &lcmd, &lbuf, &llen))
		{
			quit++; ret=-1;
		}
		else if((pcmd=='l' && lcmd!='l') || (pcmd=='L' && lcmd!='L'))
		{
			log_and_send("unexpected lcmd after link in manifest\n");
			quit++; ret=-1;
		}

		if(!quit)
		{
			// remove newline.
			if(pbuf[plen]=='\n') pbuf[plen]='\0';
			if(check_regex(regex, pbuf))
			{
				if(async_write(scmd, sbuf, slen)
				  || async_write(pcmd, pbuf, plen))
				{ quit++; ret=-1; }
				else if((pcmd=='l' || pcmd=='L')
				  && async_write(lcmd, lbuf, llen))
				{ quit++; ret=-1; }
			}
		}
		if(sbuf) { free(sbuf); sbuf=NULL; }
		if(pbuf) { free(pbuf); pbuf=NULL; }
		if(lbuf) { free(lbuf); lbuf=NULL; }
	}
	gzclose_fp(&fp);
	return ret;
}

// a = length of struct bu array
// i = position to restore from
static int restore_file(struct bu *arr, int a, int i, const char *datapth, const char *fname, const char *tmppath1, const char *tmppath2)
{
	int x=0;
	char msg[256]="";
	// Go up the array until we find the file in the data directory.
	for(x=i; x<a; x++)
	{
		char *path=NULL;
		struct stat statp;
		if(!(path=prepend_s(arr[x].data, datapth, strlen(datapth))))
		{
			log_and_send("out of memory");
			return -1;
		}

		logp("server file: %s\n", path);

		if(lstat(path, &statp) || !S_ISREG(statp.st_mode))
		{
			free(path);
			continue;
		}
		else
		{
			int patches=0;
			struct stat dstatp;
			const char *tmp=NULL;
			const char *best=NULL;

			best=path;
			tmp=tmppath1;
			// Now go down the array, applying any deltas.
			for(x-=1; x>=i; x--)
			{
				char *dpath=NULL;

				if(!(dpath=prepend_s(arr[x].delta,
						datapth, strlen(datapth))))
				{
					log_and_send("out of memory");
					free(path);
					return -1;
				}
printf("try: %s\n", dpath);

				if(lstat(dpath, &dstatp)
				  || !S_ISREG(dstatp.st_mode))
				{
					free(dpath);
					continue;
				}
printf("ok\n");

				if(!patches)
				{
					// Need to gunzip the first one.
					if(inflate_oldfile(best, tmp))
					{
						logp("error when inflating %s\n", best);
						free(path);
						free(dpath);
						return -1;
					}
					best=tmp;
					if(tmp==tmppath1) tmp=tmppath2;
					else tmp=tmppath1;
				}

				if(do_patch(best, dpath, tmp,
				  FALSE /* do not gzip the result */ ))
				{
					char msg[256]="";
					snprintf(msg, sizeof(msg),
						"error when patching %s\n",
							path);
					log_and_send(msg);
					free(path);
					free(dpath);
					return -1;
				}

				best=tmp;
				if(tmp==tmppath1) tmp=tmppath2;
				else tmp=tmppath1;
				unlink(tmp);
				patches++;
			}


			logp("sending: %s\n", best);
			async_write('f', fname, strlen(fname));
			if(patches)
			{
				// If we did some patches, the resulting file
				// is not gzipped. Gzip it during the send. 
				if(send_whole_file_gz(best, datapth))
				{
					free(path);
					return -1;
				}
			}
			else
			{
				// If we did not do some patches, the resulting
				// file is already gzipped. Send it as it is.
				if(send_whole_file(best, datapth))
				{
					free(path);
					return -1;
				}
			}
			free(path);
			if(async_write_str('x', "endfile"))
				return -1;
			return 0;
		}
	}

	snprintf(msg, sizeof(msg), "restore could not find %s (%s)\n", fname, datapth);
	log_and_send(msg);
	return -1;
}

// a = length of struct bu array
// i = position to restore from
static int restore_manifest(struct bu *arr, int a, int i, const char *tmppath1, const char *tmppath2, regex_t *regex)
{
	int ret=0;
	gzFile fp=NULL;
	char *manifest=NULL;
	char *datadir=NULL;

	if(!(manifest=prepend_s(arr[i].path, "manifest.gz", strlen("manifest.gz"))))
	{
		log_and_send("out of memory");
		ret=-1;
	}
	else if(!(fp=gzopen_file(manifest, "rb9")))
	{
		log_and_send("could not open manifest");
		ret=-1;
	}
	else
	{
		char cmd;
		int quit=0;
		size_t len=0;

		reset_filecounter();

		while(!quit)
		{
			int ars=0;
			size_t slen=0;
			char *buf=NULL;
			char  *datapth=NULL;
			char  *statbuf=NULL;
			struct stat statp;
			if(async_read_quick(&cmd, &buf, &len))
			{
printf("read quick error\n");
				ret=-1; quit++; break;
			}
			if(buf)
			{
				logp("got read quick\n");
				if(cmd=='w')
				{
					logp("WARNING: %s\n", buf);
					do_filecounter(cmd, 0);
					free(buf); buf=NULL;
					continue;
				}
				else if(cmd=='i')
				{
					// Client wanted to interrupt the
					// sending of a file. But if we are
					// here, we have already moved on.
					// Ignore.
					free(buf); buf=NULL;
					continue;
				}
				else
				{
					logp("unexpected cmd from client: %c:%s\n", cmd, buf);
					free(buf); buf=NULL;
					ret=-1; quit++; break;
				}
			}
			if((ars=async_read_stat(NULL, fp, &statbuf, &slen, &statp, &datapth)))
			{
				if(ars<0) ret=-1;
				// ars==1 means end ok
				quit++;
			}
			else if(async_read_fp(NULL, fp, &cmd, &buf, &len))
			{
				ret=-1; quit++;
			}
			else
			{
				// remove newline.
				if(buf[len]=='\n') buf[len]='\0';
				if(check_regex(regex, buf))
				{
				  logp("restore: %s\n", buf);
				  if((datapth && async_write('t',
					datapth, strlen(datapth)))
				  	  || async_write('r', statbuf, slen))
				  {
					ret=-1; quit++;
				  }
				  else if(cmd=='f')
				  {
					if(restore_file(arr, a, i, datapth,
					  buf, tmppath1, tmppath2))
					{
						ret=-1; quit++;
					}
				  }
				  else
				  {
					if(async_write(cmd, buf, len))
					{
						ret=-1; quit++;
					}
					// If it is a link, read and send what
					// it points to.
					else if(cmd=='L' || cmd=='l')
					{
						free(buf); buf=NULL;
						if(async_read_fp(NULL, fp, &cmd,
							&buf, &len)
						  || async_write(cmd, buf, len))
						{
							ret=-1; quit++;
						}
					}
				  }
				}
				else if(cmd=='L' || cmd=='l')
				{
					// Not sending a link - read and throw
					// away the next line.
					free(buf); buf=NULL;
					if(async_read_fp(NULL, fp, &cmd, &buf, &len))
					{
						ret=-1; quit++;
					}
				}
			}
			if(buf) { free(buf); buf=NULL; }
			if(datapth) { free(datapth); datapth=NULL; }
			if(statbuf) { free(statbuf); statbuf=NULL; }
		}
		gzclose_fp(&fp);

		if(!ret && async_write_str('c', "restoreend"))
		{
			ret=-1; quit++;
		}
		if(!ret) logp("wrote restoreend\n");
		if(!ret) quit=0;
		if(ret) logp("did not write restoreend (%d)\n", ret);

		while(!ret && !quit)
		{
			char *buf=NULL;
			if(async_read(&cmd, &buf, &len))
			{
				ret=-1; quit++;
			}
			else if(cmd=='c' && !strcmp(buf, "restoreend ok"))
			{
				logp("got restoreend ok\n");
				quit++;
			}
			else if(cmd=='w')
			{
				logp("WARNING: %s\n", buf);
				do_filecounter(cmd, 0);
			}
			else if(cmd=='i')
			{
				// ignore - client wanted to interrupt a file
			}
			else
			{
				logp("unexpected cmd from client at end of restore: %c:%s\n", cmd, buf);
				ret=-1; quit++;
			}
			if(buf) { free(buf); buf=NULL; }
		}

		// Server restore counter not working right yet.
		//end_filecounter_restore(1);
	}
	if(manifest) free(manifest);
	if(datadir) free(datadir);
	return ret;
}

static int compile_regex(regex_t **regex, const char *str)
{
	if(str)
	{
		if(!(*regex=(regex_t *)malloc(sizeof(regex_t)))
		  || regcomp(*regex, str, REG_EXTENDED))
		{
			log_and_send("unable to compile regex\n");
			return -1;
		}
	}
	return 0;
}

static int do_list(const char *basedir, const char *backup, const char *listregex)
{
	int a=0;
	int i=0;
	int ret=0;
	int found=0;
	struct bu *arr=NULL;
	unsigned long index=0;
	regex_t *regex=NULL;

	logp("in do_list\n");

	if(compile_regex(&regex, listregex)) return -1;

	if(get_current_backups(basedir, &arr, &a))
	{
		if(regex) { regfree(regex); free(regex); }
		return -1;
	}

	if(backup && *backup) index=strtoul(backup, NULL, 10);

	for(i=0; i<a; i++)
	{
		// Search all backups for things matching the regex.
		if(listregex && backup && *backup=='a')
		{
			found=TRUE;
			async_write('b',
				arr[i].timestamp, strlen(arr[i].timestamp));
			ret+=list_manifest(arr[i].path, regex);
		}
		// Search or list a particular backup.
		else if(backup && *backup)
		{
			if(!found
			  && (!strcmp(arr[i].timestamp, backup)
				|| arr[i].index==index))
			{
				found=TRUE;
				async_write('b',
				  arr[i].timestamp, strlen(arr[i].timestamp));
				ret=list_manifest(arr[i].path, regex);
			}
		}
		// List the backups.
		else
		{
			found=TRUE;
			async_write('b',
				arr[i].timestamp, strlen(arr[i].timestamp));
		}
	}
	free_current_backups(&arr, a);

	if(backup && *backup && !found)
	{
		async_write_str('e', "backup not found");
		ret=-1;
	}
	if(regex) { regfree(regex); free(regex); }
	return ret;
}

static int do_restore(const char *basedir, const char *backup, const char *restoreregex)
{
	int a=0;
	int i=0;
	int ret=0;
	int found=0;
	struct bu *arr=NULL;
	unsigned long index=0;
	char *tmppath1=NULL;
	char *tmppath2=NULL;
	regex_t *regex=NULL;

	logp("in do_restore\n");

	if(compile_regex(&regex, restoreregex)) return -1;

	if(!(tmppath1=prepend_s(basedir, "tmp1", strlen("tmp1")))
	  || !(tmppath2=prepend_s(basedir, "tmp2", strlen("tmp2"))))
	{
		if(tmppath1) free(tmppath1);
		if(regex) { regfree(regex); free(regex); }
		return -1;
	}

	if(get_current_backups(basedir, &arr, &a))
	{
		if(tmppath1) free(tmppath1);
		if(tmppath2) free(tmppath2);
		if(regex) { regfree(regex); free(regex); }
		return -1;
	}

	if(!(index=strtoul(backup, NULL, 10)) && a>0)
	{
		// No backup specified, do the most recent.
		ret=restore_manifest(arr, a, a-1,
			tmppath1, tmppath2, regex);
		found=TRUE;
	}
	else for(i=0; i<a; i++)
	{
		if(!strcmp(arr[i].timestamp, backup)
			|| arr[i].index==index)
		{
			found=TRUE;
			logp("got: %s\n", arr[i].path);
			ret=restore_manifest(arr, a, i,
				tmppath1, tmppath2, regex);
			break;
		}
	}

	free_current_backups(&arr, a);

	if(!found)
	{
		async_write_str('e', "backup not found");
		ret=-1;
	}
	if(tmppath1)
	{
		unlink(tmppath1);
		free(tmppath1);
	}
	if(tmppath2)
	{
		unlink(tmppath2);
		free(tmppath2);
	}
	if(regex)
	{
		regfree(regex);
		free(regex);
	}
	return ret;
}

static int check_for_rubble(const char *basedir, const char *current, const char *working, const char *currentdata, const char *finishing, struct config *cconf, const char *manifesttmp)
{
	size_t len=0;
	char realwork[256]="";
	struct stat statp;
	char *logpath=NULL;
	char *fullrealwork=NULL;
	FILE *logfp=NULL;

	// If there is a 'finishing' symlink, we need to
	// run the finish_backup stuff.
	if(!lstat(finishing, &statp))
	{
		int r=0;
		logp("Found finishing symlink - attempting to complete prior backup!\n");
		r=finish_backup(basedir,
			working,
			current,
			currentdata,
			finishing,
			cconf);
		if(!r) logp("Prior backup completed OK.\n");
		return r;
	}

	if(lstat(working, &statp))
	{
		// No working directory - that is good.
		return 0;
	}
	if(!S_ISLNK(statp.st_mode))
	{
		logp("Working directory is not a symlink.\n");
		return -1;
	}

	// The working directory has not finished being populated.
	// Check what to do.
	if((len=readlink(working, realwork, sizeof(realwork)-1))<0)
	{
		logp("Could not readlink on old working directory: %s\n", strerror(errno));
		return -1;
	}
	realwork[len]='\0';
	if(!(fullrealwork=prepend_s(basedir, realwork, strlen(realwork))))
		return -1;

	if(lstat(fullrealwork, &statp))
	{
		logp("could not lstat '%s' - something odd is going on\n",
			fullrealwork);
		free(fullrealwork);
		return -1;
	}

	// We have found an old working directory - open the log inside
	// for appending.
	if(!(logpath=prepend_s(fullrealwork, "log", strlen("log")))
	  || !(logfp=open_file(logpath, "ab")) || set_logfp(logfp))
	{
		free(fullrealwork);
		if(logpath) free(logpath);
		return -1;
	}
	free(logpath);

	logp("found old working directory: %s\n", fullrealwork);
	logp("working_dir_recovery_method: %s\n",
		cconf->working_dir_recovery_method);
	if(!strcmp(cconf->working_dir_recovery_method, "delete"))
	{
		// Try to remove it and start again.
		logp("deleting old working directory\n");
		if(recursive_delete(fullrealwork,
			NULL, TRUE /* delete files */))
		{
			logp("Old working directory is in the way.\n");
			free(fullrealwork);
			set_logfp(NULL); // fclose the logfp
			return -1;
		}
		unlink(working); // get rid of the symlink.
	}
	else if(!strcmp(cconf->working_dir_recovery_method, "use"))
	{
		// Use it as it is.
		logp("converting old working directory into the latest backup\n");
		free(fullrealwork);

		// TODO: There might be a partial file written that is not
		// yet logged to the manifest. It does no harm other than
		// taking up some disk space. Detect this and remove it.

		// Just rename the working link to be a finishing link,
		// then run this function again.
		if(do_rename(working, finishing))
		{
			set_logfp(NULL); // fclose the logfp
			return -1;
		}
		set_logfp(NULL); // fclose the logfp

		return check_for_rubble(basedir, current, working,
			currentdata, finishing, cconf, manifesttmp);
	}
	else if(!strcmp(cconf->working_dir_recovery_method, "merge"))
	{
		// Forward merge with the 'current' backup.
		FILE *mp=NULL;
		gzFile czp=NULL;
		char *cmanifest=NULL;

		logp("merging old working directory with the most recent backup\n");

		// Open the partial manifest for appending.
		// Open the previous manifest for reading.
		// Forward through the previous manifest until at the
		// place where the partial one ended, then copy the rest of
		// the old one into the new one.

		free(fullrealwork);

		if(!(mp=open_file(manifesttmp, "r+b"))
		  || !(cmanifest=prepend_s(current,
			"manifest.gz", strlen("manifest.gz"))))
		{
			set_logfp(NULL); // fclose the logfp
			return -1;
		}

		if((czp=gzopen_file(cmanifest, "rb9")))
		{
			size_t plen=0;
			size_t slen=0;
			size_t dlen=0;
			char *pbuf=NULL;
			char *dbuf=NULL;
			char *sbuf=NULL;
			char *lastpbuf=NULL;
			z_off_t offz=0;
			size_t got=0;
			size_t res=0;
			char copybuf[16000];

			// Forwarding through working manifest...
			while(1)
			{
			  if(read_next_manifest_sig_line(mp, NULL, NULL,
			    NULL, &pbuf, &sbuf, &dbuf, &plen, &slen, &dlen,
			    NULL, NULL))
			  {
				free_manbufs(NULL, &pbuf, &sbuf, &dbuf);
				close_fp(&mp);
				gzclose_fp(&czp);
				free(cmanifest);
				set_logfp(NULL); // fclose the logfp
				return -1;
			  }
			  if(!pbuf || !sbuf || !dbuf) break; // got to the end

			  if(lastpbuf) free(lastpbuf);
			  lastpbuf=strdup(pbuf);
			  free_manbufs(NULL, &pbuf, &sbuf, &dbuf);
			}
			free_manbufs(NULL, &pbuf, &sbuf, &dbuf);

			// Now we should have the latest file entry.
			// Forward through the current manifest.

			if(lastpbuf)
			{
			 logp("Last path in working manifest: '%s'\n",
				lastpbuf);
			 while(1)
			 {
			  if((offz=gztell(czp))<0)
			  {
				logp("gztell returned error\n");
				free_manbufs(NULL, &pbuf, &sbuf, &dbuf);
				close_fp(&mp);
				gzclose_fp(&czp);
				free(cmanifest);
				free(lastpbuf);
				set_logfp(NULL); // fclose the logfp
				return -1;
			  }
			  if(read_next_manifest_sig_line(NULL, czp, NULL,
			    NULL, &pbuf, &sbuf, &dbuf, &plen, &slen, &dlen,
			    NULL, NULL))
			  {
				free_manbufs(NULL, &pbuf, &sbuf, &dbuf);
				close_fp(&mp);
				gzclose_fp(&czp);
				free(cmanifest);
				free(lastpbuf);
				set_logfp(NULL); // fclose the logfp
				return -1;
			  }
			  if(pbuf && pathcmp(lastpbuf, pbuf)<0)
			  {
				// Ahead in the current manifest.
				// Get back to the start of this entry.

				logp("Ahead in the current manifest\n");
				logp("working: %s\n", lastpbuf);
				logp("current: %s\n", pbuf);
				if(gzseek(czp, offz, SEEK_SET)!=offz)
				{
					logp("error in gzseek\n");
					free_manbufs(NULL, &pbuf, &sbuf, &dbuf);
					close_fp(&mp);
					gzclose_fp(&czp);
					free(cmanifest);
					free(lastpbuf);
					set_logfp(NULL); // fclose the logfp
					return -1;
				}
				break;
			  }
			  if(!pbuf || !sbuf || !dbuf) break; // got to the end
			  free_manbufs(NULL, &pbuf, &sbuf, &dbuf);
			 }
			 free_manbufs(NULL, &pbuf, &sbuf, &dbuf);
			 free(lastpbuf);
			}
			else
			{
				logp("Did not find last path in working manifest.\n");
			}
			free_manbufs(NULL, &pbuf, &sbuf, &dbuf);

			// Write the rest of the current manifest into
			// the working one.
			logp("Filling in working manifest...\n");
			while((got=gzread(czp, copybuf, sizeof(copybuf)))>0)
			{
				res=fwrite(copybuf, 1, got, mp);
				if(res!=got)
				{
					logp("merging manifest - read %d but wrote %d\n", got, res);
					close_fp(&mp);
					gzclose_fp(&czp);
					free(cmanifest);
					set_logfp(NULL); // fclose the logfp
					return -1;
				}
			}
			close_fp(&mp);
			gzclose_fp(&czp);
			logp("Merge OK\n");
		}
		else
		{
			logp("No current backup found.\n");
			logp("Will use old working directory as the current backup.\n");
		}
		free(cmanifest);

		// Then rename the working link to be a finishing link,
		// then run this function again.
		if(do_rename(working, finishing))
		{
			set_logfp(NULL); // fclose the logfp
			return -1;
		}
		set_logfp(NULL); // fclose the logfp

		return check_for_rubble(basedir, current, working,
			currentdata, finishing, cconf, manifesttmp);
	}
	else
	{
		logp("Unknown working_dir_recovery_method: %s\n",
			cconf->working_dir_recovery_method);
		free(fullrealwork);
		set_logfp(NULL); // fclose the logfp
		return -1;
	}

	free(fullrealwork);
	set_logfp(NULL); // fclose the logfp
	return 0;
}

static int get_lock_and_clean(const char *basedir, const char *lockfile, const char *current, const char *working, const char *currentdata, const char *finishing, bool cancel, bool *gotlock, struct config *cconf, const char *manifesttmp, const char *forward)
{
	int ret=0;
	if(get_lock(lockfile))
	{
		if(cancel)
		{
			logp("another instance of client '%s' is already running,\n", cconf->cname);
			logp("or %s is not writable.\n", lockfile);
			async_write_str('e', "another instance is already running");
		}
		ret=-1;
	}
	else *gotlock=TRUE;

	if(!ret && check_for_rubble(basedir, current, working, currentdata,
		finishing, cconf, manifesttmp))
			ret=-1;

	return ret;
}

static int child(struct config *conf)
{
	int ret=0;
	char cmd;
	int quit=0;
	char msg[256]="";
	char *buf=NULL;
	size_t len=0;
	char *client=NULL;
	char *basedir=NULL;
	// Do not allow a single client to connect more than once
	char *lockfile=NULL;
	bool gotlock=FALSE;
	// The previous backup
	char *current=NULL;
	// The one we are working on
	char *working=NULL;
	// The full path to the latest backup data
	char *currentdata=NULL;
	struct config cconf;
	// where the data goes initially
	char *datadirtmp=NULL;
	// File containing timestamp of the next backup in the sequence.
	char *forward=NULL;
	// The final compressed manifest 
	char *manifest=NULL;
	// The manifest that we are working on.
	char *manifesttmp=NULL;
	// A symlink that indicates that the
	// data from the client is complete and just some work on the server 
	// is needed to finish. The 'working' symlink gets renamed to this
	// at the appropriate moment.
	char *finishing=NULL;

	reset_filecounter();

	if(authorise_client(conf, &client, &cconf) || !client || !*client)
	{
		// add an annoying delay in case they are tempted to
		// try repeatedly
		sleep(3);
		log_and_send("unable to authorise");
		ret=-1;
		quit++;
	}
	else if(!(basedir=prepend_s(conf->directory, client, strlen(client)))
	  || !(working=prepend_s(basedir, "working", strlen("working")))
	  || !(finishing=prepend_s(basedir, "finishing", strlen("finishing")))
	  || !(current=prepend_s(basedir, "current", strlen("current")))
	  || !(currentdata=prepend_s(current, "data", strlen("data")))
	  || !(manifesttmp=prepend_s(working, "manifest", strlen("manifest")))
	  || !(manifest=prepend_s(working, "manifest.gz", strlen("manifest.gz")))
	  || !(datadirtmp=prepend_s(working, "data.tmp", strlen("data.tmp")))
	  || !(forward=prepend_s(current, "forward", strlen("forward")))
	  || !(lockfile=prepend_s(basedir, "lockfile", strlen("lockfile"))))
	{
		log_and_send("out of memory");
		ret=-1;
	}
	else if(mkpath(&current)) // creates basedir, without the /current part
	{
		snprintf(msg, sizeof(msg), "could not mkpath %s", working);
		log_and_send(msg);
		ret=-1;
	}
	else if(async_read(&cmd, &buf, &len))
	{
		ret=-1;
	}
	else if(cmd=='c' && !strcmp(buf, "backup"))
	{
		if(get_lock_and_clean(basedir, lockfile, current,
			working, currentdata,
			finishing, TRUE, &gotlock, &cconf,
			manifesttmp, forward))
				ret=-1;
		else
		{
			if(buf) free(buf); buf=NULL;
			async_write_str('c', "ok");
			if(async_read(&cmd, &buf, &len))
			{
				ret=-1;
			}
			else
			{
				async_write_str('c', "ok");
				ret=do_backup(basedir, current, working,
				  currentdata, finishing, buf, &cconf,
				  manifesttmp, forward);
			}
		}
	}
	else if(cmd=='c' && !strncmp(buf, "restore ", strlen("restore ")))
	{
		if(get_lock_and_clean(basedir, lockfile, current, working,
			currentdata, finishing, TRUE, &gotlock, &cconf,
			manifesttmp, forward))
				ret=-1;
		else
		{
			char *restoreregex=NULL;
			if((restoreregex=strrchr(buf, ':')))
			{
				*restoreregex='\0';
				*restoreregex++;
			}
			async_write_str('c', "ok");
			ret=do_restore(basedir, buf+strlen("restore "),
				restoreregex);
		}
	}
	else if(cmd=='c' && !strncmp(buf, "list ", strlen("list ")))
	{
		char *listregex=NULL;
		if((listregex=strrchr(buf, ':')))
		{
			*listregex='\0';
			*listregex++;
		}
		// Allow listing without having to have no backup or restore
		// running. Though, if we can, clean up before doing the list.
		get_lock_and_clean(basedir, lockfile, current, working,
			currentdata, finishing, FALSE, &gotlock, &cconf,
			manifesttmp, forward);
		async_write_str('c', "ok");
		ret=do_list(basedir, buf+strlen("list "), listregex);
	}
	else
	{
		async_write_str('e', "unknown command");
		ret=-1;
	}
	if(buf) free(buf);

	if(client) free(client);
	if(basedir) free(basedir);
	if(current) free(current);
	if(finishing) free(finishing);
	if(working) free(working);
	if(currentdata) free(currentdata);
	if(datadirtmp) free(datadirtmp);
	if(manifest) free(manifest);
	if(manifesttmp) free(manifesttmp);
	if(forward) free(forward);
	if(lockfile)
	{
		if(gotlock) unlink(lockfile);
		free(lockfile);
	}
	return ret;
}

int server(struct config *conf, int forking)
{
	int rfd=-1;

	setup_signals();

	if((rfd=init_listen_socket(conf->port))<0) return 1;

	while(1)
	{
		int cfd=-1;
		berrno be;
		int pid_status=-1;
		socklen_t client_length=0;
		struct sockaddr_in client_name;

		client_length=sizeof(client_name);
		if((cfd=accept(rfd,
			(struct sockaddr *) &client_name,
			&client_length))==-1)
		{
			logp("accept failed: %s\n", be.bstrerror());
			return 1;
		}
		reuseaddr(cfd);

		/* clean up any children whom have exited */
		while(waitpid(-1, &pid_status, WNOHANG)>0);

		if(forking)
		{
		  /* fork off our new process to handle this request */
		  switch(fork())
		  {
			case -1:
				logp("fork failed: %s\n", strerror(errno));
				break;
			case 0:
			{
				// child
				int ret=0;
				close_fd(&rfd);
				set_non_blocking(cfd);
				if(!(ret=async_init(cfd))) ret=child(conf);
				close_fd(&cfd);
				async_free();
				logp("exit child\n");
				return ret;
			}
			default:
				// parent
				close_fd(&cfd);
				break;
		  }
		}
		else
		{
			int ret=0;
			set_non_blocking(cfd);
			if(!(ret=async_init(cfd))) ret=child(conf);
			close_fd(&cfd);
			async_free();
			close_fd(&rfd);
			return ret;
		}
	}

	return 0;
}
